import asyncio
import os
import warnings
from typing import Optional
from ..constants import (
    API_KEY_ENV,
    IMPLEMENTATION_ID_ENV,
    LOG_TO_SERVER_ENV,
    IMPLEMENTATION_ID_NAME,
)
from abc import abstractmethod
from ..client import Client
from ..utils import softmax
from ..singleton import Singleton


class Metric(metaclass=Singleton):
    # set an arbitrary minimum score that will get over-ridden later
    minimum_score: float = 0

    # Measure function signature is subject to be different - not sure
    # how applicable this is - might need a better abstraction
    @abstractmethod
    def measure(self, output, expected_output, query: Optional[str] = None):
        raise NotImplementedError

    def _get_init_values(self):
        # We use this method for sending useful metadata
        init_values = {
            param: getattr(self, param)
            for param in vars(self)
            if isinstance(getattr(self, param), (str, int, float))
        }
        return init_values

    @abstractmethod
    def is_successful(self) -> bool:
        raise NotImplementedError

    def _is_api_key_set(self):
        result = os.getenv(API_KEY_ENV) is not None
        if result is False:
            warnings.warn(
                """API key is not set. Please set it by visiting https://app.confident-ai.com"""
            )
        return result

    def _is_send_okay(self):
        # DOing this until the API endpoint is fixed
        return self._is_api_key_set() and os.getenv(LOG_TO_SERVER_ENV) != "Y"

    def __call__(self, *args, **kwargs):
        score = self.measure(*args, **kwargs)
        success = score >= self.minimum_score
        asyncio.create_task(
            self._send_to_server(
                metric_score=score,
                metric_name=self.__name__,
                query=kwargs.get("query", "-"),
                output=kwargs.get("output", "-"),
                expected_output=kwargs.get("expected_output", "-"),
                success=success,
            )
        )
        return score

    def log(
        self,
        success: bool = True,
        score: float = 1e-10,
        metric_name: str = "-",
        query: str = "-",
        output: str = "-",
        expected_output: str = "-",
        metadata: Optional[dict] = None,
    ):
        """Log to the server.

        Parameters
        - query: What was asked to the model. This can also be context.
        - output: The LLM output.
        - expected_output: The output that's expected.
        """
        if self._is_send_okay():
            asyncio.create_task(
                self._send_to_server(
                    metric_score=score,
                    metric_name=metric_name,
                    query=query,
                    output=output,
                    expected_output=expected_output,
                    success=success,
                    metadata=metadata,
                )
            )

    async def _send_to_server(
        self,
        metric_score: float,
        metric_name: str,
        query: str = "-",
        output: str = "-",
        expected_output: str = "-",
        success: Optional[bool] = None,
        metadata: Optional[dict] = None,
        **kwargs
    ):
        if self._is_send_okay():
            client = Client(api_key=os.getenv(API_KEY_ENV))
            implementation_name = os.getenv(IMPLEMENTATION_ID_NAME)
            # implementation_id = os.getenv(IMPLEMENTATION_ID_ENV, "")
            # if implementation_id != "":
            implementation_id = client.get_implementation_id_by_name(
                implementation_name
            )
            os.environ[IMPLEMENTATION_ID_ENV] = implementation_id
            datapoint_id = client.add_golden(
                query=query,
                expected_output=expected_output,
            )
            if success is None:
                success = bool(self.is_successful())
                print({"success": success, "og": self.is_successful()})

            metric_metadata: dict = self._get_init_values()
            if metadata:
                metric_metadata.update(metadata)

            return client.add_test_case(
                metric_score=float(metric_score),
                metric_name=metric_name,
                actual_output=output,
                query=query,
                implementation_id=implementation_id,
                metrics_metadata=metric_metadata,
                success=bool(success),
                datapoint_id=datapoint_id["id"],
            )


class EntailmentScoreMetric(Metric):
    def __init__(self, model_name: str = "cross-encoder/nli-deberta-base"):
        # We use a smple cross encoder model
        from sentence_transformers import CrossEncoder

        self.model = CrossEncoder(model_name)

    def measure(self, a: str, b: str):
        scores = self.model.predict([(a, b)])
        # https://huggingface.co/cross-encoder/nli-deberta-base
        # label_mapping = ["contradiction", "entailment", "neutral"]
        return softmax(scores)[0][1]



"""Test BERT score
"""
import pytest
from deepeval.metrics.bertscore_metric import BertScoreMetric


@pytest.mark.asyncio
async def test_bert_score():
    scorer = BertScoreMetric()
    score = scorer.measure("Why are you weird", "Why are you strange?")
    assert scorer.is_successful()



"""This is the ideal user flow
"""
import pytest
import os
from deepeval.api import Api
from deepeval.metrics.factual_consistency import assert_factual_consistency
from deepeval.metrics.bertscore_metric import BertScoreMetric
from deepeval.constants import API_KEY_ENV

IMPLEMENTATION_NAME = "Quickstart Example 2"
os.environ["CONFIDENT_AI_IMP_NAME"] = IMPLEMENTATION_NAME


def generate_llm_output(query: str):
    expected_output = "Our customer success phone line is 1200-231-231."
    return expected_output


@pytest.mark.asyncio
async def test_llm_output():
    query = "What is the customer success phone line?"
    expected_output = "Our customer success phone line is 1200-231-231."
    output = generate_llm_output(query)
    assert_factual_consistency(output, expected_output)


@pytest.mark.asyncio
async def test_llm_output_custom():
    query = "What is the customer success phone line?"
    expected_output = "Dogs and cats love to walk around the beach."
    output = generate_llm_output(query)
    metric = BertScoreMetric(minimum_score=0.98)
    with pytest.raises(AssertionError):
        assert_factual_consistency(output, expected_output)


def test_implementation_inside_quickstart():
    client = Api()
    imps = client.list_implementations()
    FOUND = False
    for imp in imps:
        if imp["name"] == IMPLEMENTATION_NAME:
            FOUND = True
    assert FOUND, f"Not found in {imps}"



# from .client import Evaluator
# from .query_generator import QueryGenerator
# from .test_utils import assert_exact_match, TestEvalCase
# from .bulk_runner import TestCase, BulkTestRunner
from . import dataset
from . import metrics
from . import query_generator
from . import test_case
from . import test_utils
from . import bulk_runner
from . import evaluator
from . import utils
from . import _version
from ._version import __version__



"""Overall Score
"""
import asyncio
from .metric import Metric
from .entailment_metric import EntailmentScoreMetric
from .answer_relevancy import AnswerRelevancy
from ..singleton import Singleton


class OverallScoreMetric(Metric, metaclass=Singleton):
    def __init__(self, minimum_score: float = 0.5):
        self.minimum_score = minimum_score
        self.entailment_metric = EntailmentScoreMetric()
        self.answer_relevancy = AnswerRelevancy()

    def __call__(self, query, output: str, expected_output: str, context: str):
        score = self.measure(
            query=query,
            output=output,
            expected_output=expected_output,
            context=context,
        )
        success = score > self.minimum_score
        asyncio.create_task(
            self._send_to_server(
                metric_score=score,
                metric_name=self.__name__,
                query=context,
                output=output,
                expected_output=expected_output,
                success=success,
            )
        )
        return score

    def measure(
        self, query: str, output: str, expected_output: str, context: str
    ) -> float:
        factual_consistency_score = self.entailment_metric.measure(
            context,
            output,
        )

        answer_relevancy_score = self.answer_relevancy.measure(
            query=query, output=output
        )

        answer_similarity_score = self.entailment_metric.measure(
            expected_output, output
        )

        overall_score = (
            +0.33 * factual_consistency_score
            + 0.33 * answer_relevancy_score
            + 0.33 * answer_similarity_score
        )
        self.success = bool(overall_score > self.minimum_score)
        metadata = {
            "factual_consistency": float(factual_consistency_score),
            "answer_relevancy": float(answer_relevancy_score),
            "answer_similarity_score": float(answer_similarity_score),
        }
        self.log(
            success=self.success,
            score=overall_score,
            metric_name=self.__name__,
            query=query,
            output=output,
            expected_output=output,
            metadata=metadata,
        )
        return overall_score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Alert Score"


def assert_overall_score(
    query: str,
    output: str,
    expected_output: str,
    context: str,
    minimum_score: float = 0.5,
):
    metric = OverallScoreMetric(minimum_score=minimum_score)
    score = metric.measure(
        query=query,
        output=output,
        expected_output=expected_output,
        context=context,
    )
    assert metric.is_successful(), f"Metric is not conceptually similar - got {score}"



"""Tests for answer relevancy
"""
import pytest

list1 = ["a", "b"]
list2 = ["b", "c"]
list3 = ["b", "a"]

list_dict_1 = [{"text": "a"}, {"text": "b"}]
list_dict_2 = [{"text": "b"}, {"text": "c"}]


@pytest.mark.asyncio
async def test_answer_relevancy():
    from deepeval.test_utils import assert_ranking_similarity

    assert_ranking_similarity(list1, list2, minimum_score=0.4)


@pytest.mark.asyncio
async def test_query_answer_relevancy():
    from deepeval.metrics.ranking_similarity import RankingSimilarity

    scorer = RankingSimilarity(minimum_score=0.5)
    result = scorer.measure(list_1=list1, list_2=list2)
    result_2 = scorer.measure(list_1=list1, list_2=list3)
    assert result_2 > result, "Ranking not working."


@pytest.mark.asyncio
async def test_query_answer_relevancy_dict():
    from deepeval.metrics.ranking_similarity import RankingSimilarity

    scorer = RankingSimilarity(minimum_score=0.3)
    result = scorer.measure(list_1=list_dict_1, list_2=list_dict_2)
    assert scorer.is_successful(), "Ranking dicts not working."



import asyncio
import numpy as np
from .metric import Metric
from sentence_transformers import SentenceTransformer, util


class AnswerRelevancy(Metric):
    def __init__(self, minimum_score: bool = 0.5):
        self.minimum_score = minimum_score

        # Load the model
        self.model = SentenceTransformer(
            "sentence-transformers/multi-qa-MiniLM-L6-cos-v1"
        )

    def __call__(self, query: str, output: str):
        score = self.measure(query, output)
        success = score > self.minimum_score
        if self._is_send_okay():
            asyncio.create_task(
                self._send_to_server(
                    metric_score=float(score),
                    query=query,
                    output=output,
                    metric_name=self.__name__,
                    success=bool(success),
                )
            )
        return score

    def measure(self, query, output: str) -> float:
        docs = [output]

        # Encode query and documents
        query_emb = self.model.encode(query)
        doc_emb = self.model.encode(docs)

        # Compute dot score between query and all document embeddings
        scores = util.dot_score(query_emb, doc_emb)[0].cpu().tolist()
        score = scores[0]
        self.success = score > self.minimum_score
        return score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Answer Relevancy"


def assert_answer_relevancy(query: str, output: str, minimum_score: float = 0.5):
    metric = AnswerRelevancy(minimum_score=minimum_score)
    score = metric(query=query, output=output)
    assert metric.is_successful(), (
        metric.__class__.__name__ + " was unsuccessful - " + str(score)
    )



"""Test alert score
"""

import pytest
from deepeval.metrics.alert_score import assert_alert_score
from deepeval.metrics.alert_score import AlertScoreMetric
from .utils import assert_viable_score

query = "Who won the FIFA World Cup in 2018?"
generated_text = "Winners of the FIFA world cup were the French national football team"
expected_output = "French national football team"
context = "The FIFA World Cup in 2018 was won by the French national football team. They defeated Croatia 4-2 in the final match to claim the championship."


@pytest.mark.asyncio
async def test_alert_score():
    assert_alert_score(
        query="Who won the FIFA World Cup in 2018?",
        generated_text="Winners of the FIFA world cup were the French national football team",
        expected_output="French national football team",
        context="The FIFA World Cup in 2018 was won by the French national football team. They defeated Croatia 4-2 in the final match to claim the championship.",
    )


@pytest.mark.asyncio
async def test_alert_score_metric():
    metric = AlertScoreMetric()
    score = metric.measure(
        query=query,
        generated_text=generated_text,
        expected_output=expected_output,
        context=context,
    )
    assert metric.is_successful(), "Overall score metric not working"
    assert_viable_score(score)



from typing import List, Callable
from abc import abstractmethod
from tabulate import tabulate
from .metrics.metric import Metric
from .test_case import TestCase


class BulkTestRunner:
    def __init__(self, test_filename: str = None):
        self.test_filename = test_filename

    @abstractmethod
    def bulk_test_cases(self) -> List[TestCase]:
        return []

    def run(self, completion_fn: Callable):
        table = []

        headers = [
            "Test Passed",
            "Metric Name",
            "Score",
            "Output",
            "Expected output",
            "Message",
        ]
        for case in self.bulk_test_cases:
            case: TestCase
            output = completion_fn(case.query)
            for metric in case.metrics:
                score = metric(output, case.expected_output)
                is_successful = metric.is_successful()
                message = f"""{metric.__class__.__name__} was unsuccessful for 
{case.query}
which should have matched 
{case.expected_output}
"""
                table.append(
                    [
                        is_successful,
                        metric.__class__.__name__,
                        score,
                        output,
                        case.expected_output,
                        message,
                    ]
                )
        with open(self.test_filename, "w") as f:
            f.write(tabulate(table, headers=headers))
        for t in table:
            assert t[0] == True, t[-1]
        return table



"""Tests for toxic
"""


import pytest
from deepeval.metrics.toxic_classifier import NonToxicMetric, assert_non_toxic
from .utils import assert_viable_score

generated_text = "Winners of the FIFA world cup were the French national football team"


@pytest.mark.asyncio
async def test_non_toxic():
    assert_non_toxic(
        text="Who won the FIFA World Cup in 2018?",
    )


@pytest.mark.asyncio
async def test_non_toxic_2():
    with pytest.raises(AssertionError):
        assert_non_toxic(
            text="You are stupid!",
        )


@pytest.mark.asyncio
async def test_non_toxic_metric():
    metric = NonToxicMetric()
    score = metric.measure(text=generated_text)
    assert metric.is_successful(), "Overall score metric not working"



"""Simple Retry class
"""
import time
from functools import wraps


class retry:
    def __init__(self, max_retries=3, delay=1, min_success=1):
        self.max_retries = max_retries
        self.delay = delay
        self.min_success = min_success

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self.retry_count = 0
            self.success_count = 0
            while self.retry_count < self.max_retries:
                try:
                    self.retry_count += 1
                    for _ in range(self.min_success - self.success_count):
                        func(*args, **kwargs)
                        self.success_count += 1
                    return
                except AssertionError as e:
                    print(f"Attempt {self.retry_count} failed: {str(e)}")
                except Exception as e:
                    print(f"Attempt {self.retry_count} failed: {str(e)}")

                if self.retry_count < self.max_retries:
                    print(f"Retrying in {self.delay} seconds...")
                    time.sleep(self.delay)
                else:
                    print(f"Max retries ({self.max_retries}) exceeded.")
                    if self.success_count < self.min_success:
                        raise Exception(
                            "Operation failed the required number of times."
                        )

        return wrapper



"""Tests for metric
"""
import os
from deepeval.metrics.metric import Metric
from deepeval.constants import API_KEY_ENV, LOG_TO_SERVER_ENV


class RandomMetric(Metric):
    def is_successful(self):
        return True

    def measure(self, a, b):
        return 0.5


def test_is_api_key_set():
    metric = RandomMetric()
    assert metric._is_api_key_set() == False


def test_is_api_key_set():
    os.environ[API_KEY_ENV] = "XYZFHEUIHFA"
    os.environ[LOG_TO_SERVER_ENV] = "ABCDEFGHI"
    metric = RandomMetric()
    assert metric._is_api_key_set() == True



"""Test alert score
"""

import pytest
from deepeval.api import Api
from deepeval.metrics.overall_score import assert_overall_score
from deepeval.metrics.overall_score import OverallScoreMetric
from .utils import assert_viable_score
import os

IMPLEMENTATION_NAME = "Fifa"
os.environ["CONFIDENT_AI_IMP_NAME"] = IMPLEMENTATION_NAME

query = "Who won the FIFA World Cup in 2018?"
output = "Winners of the FIFA world cup were the French national football team"
expected_output = "French national football team"
context = "The FIFA World Cup in 2018 was won by the French national football team. They defeated Croatia 4-2 in the final match to claim the championship."

client = Api()


@pytest.mark.asyncio
async def test_overall_score():
    assert_overall_score(
        query=query,
        output=output,
        expected_output=expected_output,
        context=context,
    )


@pytest.mark.asyncio
async def test_overall_score_metric():
    metric = OverallScoreMetric()
    score = metric.measure(
        query=query,
        output=output,
        expected_output=expected_output,
        context=context,
    )
    assert metric.is_successful(), "Overall score metric not working"
    assert_viable_score(score)


def test_implementation_inside_overall():
    imps = client.list_implementations()
    FOUND = False
    for imp in imps:
        if imp["name"] == IMPLEMENTATION_NAME:
            FOUND = True
    assert FOUND, f"Not found in {imps}"



from typing import Optional
from .metric import Metric
from ..singleton import Singleton
from ..utils import cosine_similarity


class BertScoreMetric(Metric, metaclass=Singleton):
    """basic implementation of BertScoreMetric"""

    def __init__(
        self,
        model_name: Optional[str] = "sentence-transformers/all-mpnet-base-v2",
        minimum_score: float = 0.7,
    ):
        from sentence_transformers import SentenceTransformer

        self.model_name = model_name
        self.model = SentenceTransformer(self.model_name).eval()
        self.minimum_score = minimum_score

    def _vectorize(self, text_a: str, text_b: str):
        vectors = self.model.encode([text_a, text_b])
        return vectors

    def measure(self, a: str, b: str):
        vectors = self._vectorize(a, b)
        self.score = cosine_similarity(vectors[0], vectors[1])
        return self.score

    def is_successful(self) -> bool:
        return self.score >= self.minimum_score

    @property
    def __name__(self):
        return "BertScore"



from sentence_transformers import CrossEncoder
from ..utils import softmax
from ..singleton import Singleton
from .metric import Metric


class EntailmentScoreMetric(Metric, metaclass=Singleton):
    def __init__(
        self,
        minimum_score: float = 0.6,
        model_name: str = "cross-encoder/nli-deberta-base",
    ):
        # We use a smple cross encoder model
        self.model = CrossEncoder(model_name)
        self.minimum_score = minimum_score

    def measure(self, a: str, b: str):
        scores = self.model.predict([(a, b)])
        # https://huggingface.co/cross-encoder/nli-deberta-base
        # label_mapping = ["contradiction", "entailment", "neutral"]
        score = softmax(scores)[0][1]
        self.success = score > self.minimum_score
        return score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Entailment"



from sentence_transformers import CrossEncoder
from ..singleton import Singleton
from ..utils import softmax, chunk_text
from .metric import Metric


class FactualConsistencyMetric(Metric, metaclass=Singleton):
    def __init__(
        self,
        minimum_score: float = 0.6,
        model_name: str = "cross-encoder/nli-deberta-v3-large",
    ):
        # We use a smple cross encoder model
        self.model = CrossEncoder(model_name)
        self.minimum_score = minimum_score

    def measure(self, output: str, context: str):
        context_list = chunk_text(context)
        max_score = 0
        for c in context_list:
            scores = self.model.predict([(context, output), (output, context)])
            # https://huggingface.co/cross-encoder/nli-deberta-base
            # label_mapping = ["contradiction", "entailment", "neutral"]
            softmax_scores = softmax(scores)
            score = softmax_scores[0][1]
            if score > max_score:
                max_score = score

            second_score = softmax_scores[1][1]
            if second_score > max_score:
                max_score = second_score

        self.success = max_score > self.minimum_score
        print({"success": self.success, "score": max_score})
        self.log(
            success=self.success,
            score=max_score,
            metric_name="Factual Consistency",
            output=output,
            expected_output=context,
        )
        return max_score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Factual Consistency"


def assert_factual_consistency(output: str, context: str, minimum_score: float = 0.3):
    """Assert that the output is factually consistent with the context."""

    metric = FactualConsistencyMetric(minimum_score=minimum_score)
    score = metric(context, output)
    assert metric.is_successful(), metric.__class__.__name__ + " was unsuccessful."



import hashlib
from typing import Any, List
from collections import UserList
from .metrics.metric import Metric
from .metrics.randomscore import RandomMetric
from .metrics.factual_consistency import FactualConsistencyMetric


class TestCase:
    def __init__(
        self,
        query: Any,
        expected_output: Any,
        metrics: List[Metric] = None,
        id: str = None,
    ):
        if metrics is None:
            fact_consistency_metric = FactualConsistencyMetric(minimum_score=0.3)
            self.metrics = [fact_consistency_metric]
        else:
            self.metrics = metrics
        self.query = query
        self.expected_output = expected_output
        if id is None:
            id_string = str(self.query) + str(self.expected_output)
            self.id = hashlib.md5(id_string.encode()).hexdigest()
        else:
            self.id = id

    def dict(self):
        return {
            "query": self.query,
            "expected_output": self.expected_output,
            "metrics": self.metrics,
            "id": self.id,
        }


class TestCases(UserList):
    """A list of test cases for easy access"""

    @classmethod
    def from_csv(
        self,
        csv_filename: str,
        query_column: str,
        expected_output_column: str,
        id_column: str = None,
        metrics: List[Metric] = None,
    ):
        import pandas as pd

        df = pd.read_csv(csv_filename)
        querys = df[query_column].values
        expected_outputs = df[expected_output_column].values
        if id_column is not None:
            ids = df[id_column].values
        for i, query in enumerate(querys):
            self.data.append(
                TestCase(
                    query=query,
                    expected_output=expected_outputs[i],
                    metrics=metrics,
                    id=ids[i],
                )
            )
        return self



"""Test for custom metrics in Python
"""

import pytest
import asyncio
from deepeval.metrics.metric import Metric


class LengthMetric(Metric):
    """This metric checks if the output is more than 3 letters"""

    def __init__(self, minimum_length: int = 3):
        self.minimum_length = minimum_length

    def measure(self, text: str):
        # sends to server
        score = len(text)
        self.success = score > self.minimum_length
        # Optional: Logs it to the server
        self.log(
            query=text,
            score=score
            / 100,  # just to have something here - should be between 0 and 1
            success=self.success,
        )
        return score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Length"


@pytest.mark.asyncio
async def test_length_metric():
    metric = LengthMetric()
    metric.measure("fehusihfe wuifheuiwh")
    assert metric.is_successful()



"""Test clients
"""
import pytest
import unittest
from deepeval.evaluator import Evaluator
from deepeval.pipeline import Pipeline
from deepeval.metrics.randomscore import RandomMetric


@pytest.mark.skip(reason="Evaluator not finished")
class TestEvaluator(unittest.TestCase):
    def test_eval(self):
        eval = Evaluator()
        result = eval.add_ground_truth(
            query="Example",
            expected_response="Customer success response is here",
            tags=["sample"],
        )
        assert True


def result_function(query):
    return "Custiomers are great!"


@pytest.mark.skip(reason="Evaluator not finished")
class TestE2E(unittest.TestCase):
    def test_e2e(self):
        eval = Evaluator()
        pipeline = Pipeline(
            pipeline_id="textract-pdf-processing", result_function=result_function
        )
        metric = RandomMetric()
        result = eval.evaluate(pipeline=pipeline, metric=metric)
        assert True



import pytest
import os
from deepeval.metrics.factual_consistency import (
    assert_factual_consistency,
    FactualConsistencyMetric,
)

IMPLEMENTATION_NAME = "Test Factual Consistency"
os.environ["CONFIDENT_AI_IMP_NAME"] = IMPLEMENTATION_NAME


@pytest.mark.asyncio
async def test_factual_consistency():
    with pytest.raises(AssertionError):
        assert_factual_consistency(
            context="After a long day at work, Sarah decided to go for a walk in the park to unwind. She put on her sneakers and grabbed her headphones before heading out. As she strolled along the path, she noticed families having picnics, children playing on the playground, and ducks swimming in the pond.",
            output="Sarah spent the evening at the library, engrossed in a book.",
        )


@pytest.mark.asyncio
async def test_factual_consistency_2():
    assert_factual_consistency(
        context="After a long day at work, Sarah decided to go for a walk in the park to unwind. She put on her sneakers and grabbed her headphones before heading out. As she strolled along the path, she noticed families having picnics, children playing on the playground, and ducks swimming in the pond.",
        output="Sarah went out for a walk in the park.",
    )


@pytest.mark.asyncio
async def test_factual_consistency_metric():

    metric = FactualConsistencyMetric(minimum_score=0.6)
    result = metric.measure(
        output="Python is a programming language.",
        context="Python is a high-level, versatile, and interpreted programming language known for its simplicity and readability.",
    )
    assert metric.is_successful()
    assert result > 0.8, f"Failed {result}"


@pytest.mark.asyncio
async def test_factual_consistency_metric_2():

    metric = FactualConsistencyMetric(minimum_score=0.6)
    result = metric.measure(
        output="Python is a programming language.",
        context="Python is NOT a programming language.",
    )
    assert not metric.is_successful()


@pytest.mark.asyncio
async def test_factual_consistency_metric_3():

    metric = FactualConsistencyMetric(minimum_score=0.6)
    result = metric.measure(
        output="Python is a programming language.",
        context="Python is a snake.",
    )
    assert not metric.is_successful()



import functools
import time
from typing import Any, List
from .metrics.randomscore import RandomMetric
from .metrics.metric import Metric
from .metrics.bertscore_metric import BertScoreMetric
from .metrics.entailment_metric import EntailmentScoreMetric
from .metrics.ranking_similarity import RankingSimilarity


def assert_llm_output(
    output: Any, expected_output: Any, metric: Any = "entailment", query: str = "-"
):
    if metric == "exact":
        assert_exact_match(output, expected_output)
    elif metric == "random":
        metric: RandomMetric = RandomMetric()
        metric(output, expected_output, query=query)
    elif metric == "bertscore":
        metric: BertScoreMetric = BertScoreMetric()
        metric(output, expected_output, query=query)
    elif metric == "entailment":
        metric: EntailmentScoreMetric = EntailmentScoreMetric()
        metric(output, expected_output, query=query)
    elif isinstance(metric, Metric):
        metric(output, expected_output, query=query)
    else:
        raise ValueError("Inappropriate metric")
    assert metric.is_successful(), metric.__class__.__name__ + " was unsuccessful."


def assert_exact_match(text_input: str, text_output: str):
    assert text_input == text_output, f"{text_output} != {text_input}"


def assert_ranking_similarity(
    list_1: List[Any], list_2: List[Any], minimum_score: float = 0.1
):
    metric = RankingSimilarity(minimum_score=minimum_score)
    result = metric(list_1, list_2)
    assert metric.is_successful(), (
        metric.__class__.__name__ + " was unsuccessful - " + str(result)
    )


class TestEvalCase:
    pass


def timing_decorator(func):
    @functools.wraps(func)  # Preserves the original function's metadata
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        print(f"{func.__name__} took {end_time - start_time:.4f} seconds to execute.")
        return result

    return wrapper


def tags(tags: list):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            # TODO - add logging for tags
            print(f"Tags are: {tags}")
            return result

        return wrapper

    return decorator



"""Alert Score
"""
from .metric import Metric
from .entailment_metric import EntailmentScoreMetric

# from .answer_relevancy import AnswerRelevancy


class AlertScoreMetric(Metric):
    def __init__(self, minimum_score: float = 0.5):
        self.minimum_score = minimum_score
        self.entailment_metric = EntailmentScoreMetric()
        # self.answer_relevancy = AnswerRelevancy()

    def __call__(self, generated_text: str, expected_output: str, context: str):
        score = self.measure(generated_text, expected_output, context)
        return score

    def measure(
        self, query: str, generated_text: str, expected_output: str, context: str
    ) -> float:

        entailment_score = self.entailment_metric.measure(
            context,
            generated_text,
        )

        answer_expected_score = self.entailment_metric.measure(
            generated_text,
            expected_output,
        )

        # This metric is very very bad right now as it requires the answer
        # to re-gurgitate the question.
        # answer_relevancy_score = self.answer_relevancy.measure(
        #     query=query, answer=generated_text
        # )
        alert_score = min(entailment_score, answer_expected_score)
        self.success = alert_score > self.minimum_score
        return alert_score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Alert Score"


def assert_alert_score(
    query: str,
    generated_text: str,
    expected_output: str,
    context: str,
    minimum_score: float = 0.5,
):
    """Create alert score."""
    metric = AlertScoreMetric(minimum_score=minimum_score)
    score = metric.measure(
        query=query,
        generated_text=generated_text,
        expected_output=expected_output,
        context=context,
    )
    assert metric.is_successful(), f"Found issue - Alert score: {score}"



# Testing for ranking similarity
import numpy as np
import asyncio
from typing import List, Optional, Union, Any
from tqdm import tqdm
from .metric import Metric


class RBO:
    """
    This class will include some similarity measures between two different
    ranked lists.
    """

    def __init__(
        self,
        S: Union[List, np.ndarray],
        T: Union[List, np.ndarray],
        verbose: bool = False,
    ) -> None:
        """
        Initialize the object with the required lists.
        Examples of lists:
        S = ["a", "b", "c", "d", "e"]
        T = ["b", "a", 1, "d"]

        Both lists reflect the ranking of the items of interest, for example,
        list S tells us that item "a" is ranked first, "b" is ranked second,
        etc.

        Args:
            S, T (list or numpy array): lists with alphanumeric elements. They
                could be of different lengths. Both of the them should be
                ranked, i.e., each element"s position reflects its respective
                ranking in the list. Also we will require that there is no
                duplicate element in each list.
            verbose: If True, print out intermediate results. Default to False.
        """

        assert type(S) in [list, np.ndarray]
        assert type(T) in [list, np.ndarray]

        assert len(S) == len(set(S))
        assert len(T) == len(set(T))

        self.S, self.T = S, T
        self.N_S, self.N_T = len(S), len(T)
        self.verbose = verbose
        self.p = 0.5  # just a place holder

    def assert_p(self, p: float) -> None:
        """Make sure p is between (0, 1), if so, assign it to self.p.

        Args:
            p (float): The value p.
        """
        assert 0.0 < p < 1.0, "p must be between (0, 1)"
        self.p = p

    def _bound_range(self, value: float) -> float:
        """Bounds the value to [0.0, 1.0]."""

        try:
            assert 0 <= value <= 1 or np.isclose(1, value)
            return value

        except AssertionError:
            print("Value out of [0, 1] bound, will bound it.")
            larger_than_zero = max(0.0, value)
            less_than_one = min(1.0, larger_than_zero)
            return less_than_one

    def rbo(
        self,
        k: Optional[float] = None,
        p: float = 1.0,
        ext: bool = False,
    ) -> float:
        """
        This the weighted non-conjoint measures, namely, rank-biased overlap.
        Unlike Kendall tau which is correlation based, this is intersection
        based.
        The implementation if from Eq. (4) or Eq. (7) (for p != 1) from the
        RBO paper: http://www.williamwebber.com/research/papers/wmz10_tois.pdf

        If p = 1, it returns to the un-bounded set-intersection overlap,
        according to Fagin et al.
        https://researcher.watson.ibm.com/researcher/files/us-fagin/topk.pdf

        The fig. 5 in that RBO paper can be used as test case.
        Note there the choice of p is of great importance, since it
        essentially control the "top-weightness". Simply put, to an extreme,
        a small p value will only consider first few items, whereas a larger p
        value will consider more items. See Eq. (21) for quantitative measure.

        Args:
            k: The depth of evaluation.
            p: Weight of each agreement at depth d:
                p**(d-1). When set to 1.0, there is no weight, the rbo returns
                to average overlap.
            ext: If True, we will extrapolate the rbo, as in Eq. (23).

        Returns:
            The rbo at depth k (or extrapolated beyond).
        """

        if not self.N_S and not self.N_T:
            return 1  # both lists are empty

        if not self.N_S or not self.N_T:
            return 0  # one list empty, one non-empty

        if k is None:
            k = float("inf")
        k = min(self.N_S, self.N_T, k)

        # initialize the agreement and average overlap arrays
        A, AO = [0] * k, [0] * k
        if p == 1.0:
            weights = [1.0 for _ in range(k)]
        else:
            self.assert_p(p)
            weights = [1.0 * (1 - p) * p**d for d in range(k)]

        # using dict for O(1) look up
        S_running, T_running = {self.S[0]: True}, {self.T[0]: True}
        A[0] = 1 if self.S[0] == self.T[0] else 0
        AO[0] = weights[0] if self.S[0] == self.T[0] else 0

        for d in tqdm(range(1, k), disable=~self.verbose):

            tmp = 0
            # if the new item from S is in T already
            if self.S[d] in T_running:
                tmp += 1
            # if the new item from T is in S already
            if self.T[d] in S_running:
                tmp += 1
            # if the new items are the same, which also means the previous
            # two cases did not happen
            if self.S[d] == self.T[d]:
                tmp += 1

            # update the agreement array
            A[d] = 1.0 * ((A[d - 1] * d) + tmp) / (d + 1)

            # update the average overlap array
            if p == 1.0:
                AO[d] = ((AO[d - 1] * d) + A[d]) / (d + 1)
            else:  # weighted average
                AO[d] = AO[d - 1] + weights[d] * A[d]

            # add the new item to the running set (dict)
            S_running[self.S[d]] = True
            T_running[self.T[d]] = True

        if ext and p < 1:
            return self._bound_range(AO[-1] + A[-1] * p**k)

        return self._bound_range(AO[-1])


class RankingSimilarity(Metric):
    def __init__(self, minimum_score: float = 0.1):
        self.minimum_score = minimum_score

    def __call__(self, list_1: List[Any], list_2: List[Any]):
        score = self.measure(list_1, list_2)
        if self._is_send_okay():
            asyncio.create_task(
                self._send_to_server(
                    metric_score=score,
                    metric_name=self.__name__,
                    query=str(list_1),
                    output=str(list_2),
                )
            )
        return score

    def measure(self, list_1: List, list_2: List):
        list_1 = [str(x) for x in list_1]
        list_2 = [str(x) for x in list_2]
        scorer = RBO(list_1, list_2)
        result = scorer.rbo(p=0.9, ext=True)
        self.success = result > self.minimum_score
        return result

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Ranking Similarity"



"""Tests for toxic
"""


import pytest
from deepeval.metrics.bias_classifier import UnBiasedMetric, assert_unbiased

generated_text = "Winners of the FIFA world cup were the French national football team"


@pytest.mark.asyncio
async def test_non_toxic():
    assert_unbiased(
        text="Who won the FIFA World Cup in 2018?",
    )


@pytest.mark.asyncio
async def test_non_toxic_2():
    with pytest.raises(AssertionError):
        assert_unbiased(
            text="Left wing is evil.",
        )


@pytest.mark.asyncio
async def test_non_toxic_metric():
    metric = UnBiasedMetric()
    score = metric.measure(text=generated_text)
    assert metric.is_successful(), "Overall score metric not working"



"""Metric for toxic classifier. 
1 - Healthy
0 - Toxic
"""
import asyncio
from .metric import Metric
from ..singleton import Singleton


class NonToxicMetric(Metric, metaclass=Singleton):
    def __init__(self, model_name: str = "original", minimum_score: float = 0.5):
        from detoxify import Detoxify

        self.model_name = model_name
        self.model = Detoxify(model_name)
        self.minimum_score = minimum_score

    def __call__(self, text: str):
        score = self.measure(text)
        score = score["min_score"]
        if self._is_send_okay():
            asyncio.create_task(
                self._send_to_server(
                    metric_score=score,
                    metric_name=self.__name__,
                    query=text,
                    output="-",
                )
            )

    def measure(self, text: str):
        results = self.model.predict(text)
        # sample output
        # {'toxicity': 0.98057544,
        # 'severe_toxicity': 0.106649496,
        # 'obscene': 0.32923067,
        # 'threat': 0.018646086,
        # 'insult': 0.7514423,
        # 'identity_attack': 0.86643445}
        self.success = True
        for i, (k, v) in enumerate(results.items()):
            if i == 0:
                min_score = 1 - v
            if v > 1 - self.minimum_score:
                self.success = False
                min_score = 1 - v
        results["min_score"] = min_score
        return results

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Toxicness'"


def assert_non_toxic(text: str, minimum_score: float = 0.5):
    metric = NonToxicMetric(minimum_score=minimum_score)
    score = metric.measure(text)
    assert metric.is_successful(), f"Text is toxic - got {score}"



"""Client for Twilix Evals
"""
import os
import asyncio
import getpass
from tqdm.auto import tqdm
from .pipeline import Pipeline
from .client import Client
from .metrics.metric import Metric
from .query_generator import BEIRQueryGenerator, QueryGenerator
from typing import Optional, List, Dict


class Evaluator(Client):
    def __init__(self, api_key: str = None, local_mode: bool = False, **kwargs):
        if api_key is None:
            if "TWILIX_API_KEY" not in os.environ:
                api_key = getpass.getpass(
                    "Grab your API key from https://app.twilix.io"
                )
            else:
                api_key = os.environ["TWILIX_API_KEY"]
        self.api_key = api_key
        self.local_mode = local_mode
        if self.local_mode:
            self.data = []
        super().__init__(api_key=api_key, **kwargs)

    def add_ground_truth(self, query: str, expected_response: str, tags: list = None):
        """Add ground truth"""
        if self.local_mode:
            self.data.append(
                {"query": query, "expected_response": expected_response, "tags": tags}
            )
        return self.post_request(
            endpoint="v1/pipeline/eval",
            body={"query": query, "expected_response": expected_response, "tags": tags},
        )

    async def add_ground_truth_async(
        self, query: str, expected_response: str, tags: list = None
    ):
        """Add ground truth"""
        return self.post_request(
            endpoint="v1/pipeline/eval",
            body={"query": query, "expected_response": expected_response, "tags": tags},
        )

    def _get_ground_truths(self, take: int = 50, skip: int = 0):
        return self.get_request(
            endpoint="v1/pipeline/eval", params={"take": take, "skip": skip}
        )

    def _post_evaluation_result(
        self, evaluation_score: float, pipeline_id: str, test_id: str = None
    ):
        return self.post_request(
            endpoint="v1/pipeline/eval/run",
            body={
                "testId": test_id,
                "evaluationScore": evaluation_score,
                "pipelineGroupName": pipeline_id,
            },
        )

    def evaluate(self, pipeline: Pipeline, metric: Metric, test_id: str = None):
        #  {
        #     "created_at": "2023-08-10T09:28:44.782Z",
        #     "query": "Example",
        #     "expected_response": "Customer success response is here",
        #     "Tags": [
        #         "sample"
        #     ],
        #     "EvalRuns": []
        # },
        truths = self._get_ground_truths()
        for t in tqdm(truths["runs"]):
            result = pipeline.result_function(t["query"])
            expected_result = t["expected_response"]
            score = metric.measure(result, expected_result)
            result = self._post_evaluation_result(
                evaluation_score=score,
                pipeline_id=pipeline.pipeilne_id,
                test_id=t["id"],
            )
            print(result)

    async def generate_queries_async(
        self,
        texts: List,
        tags: Optional[List] = None,
        num_queries: int = 3,
        generator: QueryGenerator = None,
        max_workers: int = 3,  # Maximum number of threads and blocks until it's free
    ):
        """Generate these queries"""
        if generator is None:
            generator = BEIRQueryGenerator()

        semaphore = asyncio.Semaphore(max_workers)

        for i, t in enumerate(tqdm(texts)):
            async with semaphore:
                results = generator.generate_queries(
                    texts=[t], tags=tags, num_queries=num_queries
                )
                for j, r in enumerate(results):
                    asyncio.create_task(
                        self.add_ground_truth_async(
                            query=r, expected_response=t, tags=[tags[i]]
                        )
                    )
        assert len(results) > 0, "Provide some results."

    def generate_queries(
        self,
        texts: List,
        tags: Optional[List] = None,
        num_queries: int = 3,
        generator: QueryGenerator = None,
        max_workers: int = 3,
    ):
        asyncio.run(
            self.generate_queries_async(
                texts=texts,
                tags=tags,
                num_queries=num_queries,
                generator=generator,
                max_workers=max_workers,
            )
        )

    def get_implementation_id_by_name(
        self, name: str, description: Optional[str] = None
    ):
        """Gets implementation. If none exists, it creates one."""
        implementations: List[Dict] = self.list_implementations()
        imp_id = None
        for imp in implementations:
            if imp["name"] == name:
                imp_id = imp["id"]
        if imp_id is None:
            created_imp = self.create_implementation(name=name, description=description)
            imp_id = created_imp["id"]
        return imp_id



"""Tests for answer relevancy
"""
import pytest

query = "What is Python?"
answer = "Python is a programming language?"


@pytest.mark.asyncio
async def test_answer_relevancy():
    from deepeval.metrics.answer_relevancy import assert_answer_relevancy

    assert_answer_relevancy(query, answer, minimum_score=0.5)


@pytest.mark.asyncio
async def test_query_answer_relevancy():
    from deepeval.metrics.answer_relevancy import AnswerRelevancy

    scorer = AnswerRelevancy(minimum_score=0.5)
    result = scorer.measure(query=query, output=answer)



"""Class for Evaluation Datasets
"""
import json
import random
from tabulate import tabulate
from datetime import datetime
from typing import List, Callable
from collections import UserList
from .test_case import TestCase
from .metrics.metric import Metric
from .query_generator import BEIRQueryGenerator
from .retry import retry


class EvaluationDataset(UserList):
    """Class for Evaluation Datasets -  which are a list of test cases"""

    def __init__(self, test_cases: List[TestCase]):
        self.data: List[TestCase] = test_cases

    @classmethod
    def from_csv(
        cls,  # Use 'cls' instead of 'self' for class methods
        csv_filename: str,
        query_column: str,
        expected_output_column: str,
        id_column: str = None,
        metrics: List[Metric] = None,
    ):
        import pandas as pd

        df = pd.read_csv(csv_filename)
        querys = df[query_column].values
        expected_outputs = df[expected_output_column].values
        if id_column is not None:
            ids = df[id_column].values

        # Initialize the 'data' attribute as an empty list
        cls.data = []

        for i, query_data in enumerate(querys):
            cls.data.append(
                TestCase(
                    query=query_data,
                    expected_output=expected_outputs[i],
                    metrics=metrics,
                    id=ids[i] if id_column else None,
                )
            )
        return cls(cls.data)

    def from_test_cases(self, test_cases: list):
        self.data = test_cases

    @classmethod
    def from_json(
        cls,
        json_filename: str,
        query_column: str,
        expected_output_column: str,
        id_column: str = None,
        metrics: List[Metric] = None,
    ):
        """
        This is for JSON data in the format of key-value array pairs.
        {
            "query": ["What is the customer success number", "What is the customer success number"],
        }

        if the JSON data is in a list of dicionaries, use from_json_list
        """
        with open(json_filename, "r") as f:
            data = json.load(f)
        test_cases = []

        for i, query in enumerate(data[query_column]):
            test_cases.append(
                TestCase(
                    query=data[query_column][i],
                    expected_output=data[expected_output_column][i],
                    metrics=metrics,
                    id=data[id_column][i],
                )
            )
        return cls(data)

    @classmethod
    def from_json_list(
        cls,
        json_filename: str,
        query_column: str,
        expected_output_column: str,
        id_column: str = None,
        metrics: List[Metric] = None,
    ):
        """
        This is for JSON data in the format of a list of dictionaries.
        [
            {"query": "What is the customer success number", "expected_output": "What is the customer success number"},
        ]
        """
        with open(json_filename, "r") as f:
            data = json.load(f)
        test_cases = []
        for i, query in enumerate(data):
            test_cases.append(
                TestCase(
                    query=data[i][query_column],
                    expected_output=data[i][expected_output_column],
                    metrics=metrics,
                    id=data[i][id_column],
                )
            )
        return cls(data)

    @classmethod
    def from_dict(
        cls,
        data: List[dict],
        query_key: str,
        expected_output_key: str,
        id_key: str = None,
        metrics: List[Metric] = None,
    ):
        pass

    def to_dict(self):
        return [x.dict() for x in self.data]

    def to_csv(self, csv_filename: str):
        import pandas as pd

        df = pd.DataFrame(self.data)
        df.to_csv(csv_filename, index=False)

    def to_json(self, json_filename: str):
        with open(json_filename, "w") as f:
            json.dump(self.data, f)

    def from_hf_evals(self):
        raise NotImplementedError

    def from_df(self):
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data})"

    def sample(self, n: int = 5):
        if len(self.data) <= n:
            n = len(self.data)
        result = random.sample(self.data, n)
        return [r.dict() for r in result]

    def __getitem__(self, index):
        return self.data[index]

    def __setitem__(self, index, value):
        self.data[index] = value

    def __delitem__(self, index):
        del self.data[index]

    def run_evaluation(
        self,
        completion_fn: Callable,
        test_filename: str = None,
        max_retries: int = 3,
        min_success: int = 1,
        raise_error_on_run: bool = False,
        metrics: List[TestCase] = None,
    ):
        """Run evaluation with given metrics"""
        table = []

        headers = [
            "Test Passed",
            "Metric Name",
            "Score",
            "Output",
            "Expected output",
            "Message",
        ]
        for case in self.data:
            case: TestCase
            output = completion_fn(case.query)
            if metrics is None:
                metrics = case.metrics
            for metric in metrics:

                @retry(max_retries=max_retries, min_success=min_success)
                def assert_metric():
                    score = metric(output, case.expected_output)
                    is_successful: bool = metric.is_successful()
                    is_successful: bool = bool(is_successful)
                    message = f"""{metric.__class__.__name__} was unsuccessful for 
{case.query} 
which should have matched 
{case.expected_output}
"""
                    table.append(
                        [
                            is_successful,
                            metric.__class__.__name__,
                            score,
                            output,
                            case.expected_output,
                            message,
                        ]
                    )
                    assert is_successful, metric.__name__ + " wasn't successful"

                if raise_error_on_run:
                    assert_metric()
                else:
                    try:
                        assert_metric()
                    except AssertionError as e:
                        print(e)
                    except Exception as e:
                        print(e)
        if test_filename is None:
            test_filename = (
                f"test-result-{datetime.now().__str__().replace(' ', '-')}.txt"
            )
        with open(test_filename, "w") as f:
            f.write(tabulate(table, headers=headers))
        print(f"Saved to {test_filename}")
        for t in table:
            assert t[0] == True, t[-1]
        return test_filename


def create_evaluation_dataset_from_raw_text(text: str, output_fn: str = "output.csv"):
    """Utility function to create an evaluation dataset from raw text."""
    print(f"Outputting to {output_fn}")

    # NOTE: loading this may take a while as the model used is quite big
    gen = BEIRQueryGenerator()
    text = "Synthetic queries are useful for scenraios where there is no data."
    queries = gen.generate_queries(texts=[text], num_queries=2)
    test_cases = []
    with open(output_fn, "w") as f:
        f.write("query,expected_output\n")
        for query in queries:
            f.write(f"{query}, {text}\n")
        test_case = TestCase(query=text, expected_output=text)
        test_cases.append(test_case)

    dataset = EvaluationDataset(test_cases=test_cases)
    return dataset



import platform
import urllib.parse
import os
import requests
from .constants import API_KEY_ENV
from typing import Any, Optional
from requests.adapters import HTTPAdapter, Response, Retry

API_BASE_URL = "https://app.confident-ai.com/api"

# Parameters for HTTP retry
HTTP_TOTAL_RETRIES = 3  # Number of total retries
HTTP_RETRY_BACKOFF_FACTOR = 2  # Wait 1, 2, 4 seconds between retries
HTTP_STATUS_FORCE_LIST = [408, 429] + list(range(500, 531))
HTTP_RETRY_ALLOWED_METHODS = frozenset({"GET", "POST", "DELETE"})


class Api:
    """Internal Api reference for handling http operations"""

    def __init__(
        self,
        api_key: str = os.getenv(API_KEY_ENV),
        user_agent_extension=None,
        api_instance_url=None,
        verify=None,
        proxies=None,
        cert=None,
    ):
        if api_key == "" or api_key is None:
            raise ValueError("Please provide a valid API Key.")

        self.api_key = api_key

        self._auth = (self.api_key, "")
        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": self._generate_useragent(user_agent_extension),
            # "Authorization": "Bearer " + api_key,
            # This is what gets sent now - auth gets sent to firebase instead
            "CONFIDENT_API_KEY": api_key,
        }
        self._headers_multipart_form_data = {
            "User-Agent": self._generate_useragent(user_agent_extension),
        }
        self.base_api_url = api_instance_url or API_BASE_URL

        self.verify = verify
        self.proxies = proxies
        self.cert = cert

    @staticmethod
    def _http_request(
        method,
        url,
        headers=None,
        auth=None,
        params=None,
        body=None,
        files=None,
        data=None,
        verify=None,
        proxies=None,
        cert=None,
    ) -> Response:
        https = requests.Session()
        retry_strategy = Retry(
            total=HTTP_TOTAL_RETRIES,
            backoff_factor=HTTP_RETRY_BACKOFF_FACTOR,
            status_forcelist=HTTP_STATUS_FORCE_LIST,
            allowed_methods=HTTP_RETRY_ALLOWED_METHODS,
            raise_on_status=False,
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)
        https.mount("https://", adapter)

        https.cert = cert if cert else None
        https.verify = verify if verify else True
        if proxies:
            https.proxies.update(proxies)

        try:
            params = params or {}
            body = body or None

            res = https.request(
                method=method,
                url=url,
                headers=headers,
                # auth=auth,
                params=params,
                json=body,
                files=files,
                data=data,
            )

            return res
        except Exception as err:
            raise Exception(err) from err

    @staticmethod
    def _raise_on_response(res: Response):
        try:
            message = res.json().get("error", res.text)
        except ValueError:
            message = res.text
        raise Exception(message)

    def _api_request(
        self,
        method,
        endpoint,
        headers=None,
        auth=None,
        params=None,
        body=None,
        files=None,
        data=None,
    ):
        """Generic HTTP request method with error handling."""

        url = f"{self.base_api_url}/{endpoint}"
        res = self._http_request(
            method,
            url,
            headers,
            auth,
            params,
            body,
            files,
            data,
            verify=self.verify,
            proxies=self.proxies,
            cert=self.cert,
        )

        json = None
        if res.status_code == 200:
            try:
                json = res.json()
            except ValueError:
                # Some endpoints only return 'OK' message without JSON
                return json
        elif res.status_code == 409 and "task" in endpoint and body.get("unique_id"):
            retry_history = res.raw.retries.history
            # Example RequestHistory tuple
            # RequestHistory(method='POST',
            #   url='/v1/task/imageannotation',
            #   error=None,
            #   status=409,
            #   redirect_location=None)
            if retry_history != ():
                # See if the first retry was a 500 or 503 error
                if retry_history[0][3] >= 500:
                    uuid = body["unique_id"]
                    newUrl = f"{self.base_api_url}/tasks?unique_id={uuid}"
                    # grab task from api
                    newRes = self._http_request(
                        "GET", newUrl, headers=headers, auth=auth
                    )
                    json = newRes.json()["docs"][0]
            else:
                self._raise_on_response(res)
        else:
            self._raise_on_response(res)
        return json

    def get_request(self, endpoint, params=None):
        """Generic GET Request Wrapper"""
        return self._api_request(
            "GET", endpoint, headers=self._headers, auth=self._auth, params=params
        )

    def post_request(self, endpoint, body=None, files=None, data=None):
        """Generic POST Request Wrapper"""
        return self._api_request(
            "POST",
            endpoint,
            headers=self._headers
            if files is None
            else self._headers_multipart_form_data,
            auth=self._auth,
            body=body,
            files=files,
            data=data,
        )

    def delete_request(self, endpoint, params=None, body=None):
        """Generic DELETE Request Wrapper"""
        return self._api_request(
            "DELETE",
            endpoint,
            headers=self._headers,
            auth=self._auth,
            params=params,
            body=body,
        )

    def put_request(self, endpoint, body=None, params=None):
        """Generic PUT Request Wrapper"""
        return self._api_request(
            "PUT",
            endpoint,
            body=body,
            headers=self._headers,
            auth=self._auth,
            params=params,
        )

    @staticmethod
    def _generate_useragent(extension: str = None) -> str:
        """Generates UserAgent parameter with module, Python
        and OS details

        Args:
            extension (str, optional): Option to extend UserAgent
            with source system

        Returns:
            str: Generated UserAgent parameter with platform versions
        """
        python_version = platform.python_version()
        os_platform = platform.platform()

        user_agent = " ".join(
            filter(
                None,
                [
                    f"Python/{python_version}",
                    f"OS/{os_platform}",
                    extension,
                ],
            )
        )
        return user_agent

    @staticmethod
    def quote_string(text: str) -> str:
        """Project and Batch names can be a part of URL, which causes
        an error in case of a special character used.
        Quotation assures the right object to be retrieved from API.

        `quote_string('a bc/def')` -> `a%20bc%2Fdef`

        Args:
            text (str): Input text to be quoted

        Returns:
            str: Quoted text in return
        """
        return urllib.parse.quote(text, safe="")

    def add_test_case(
        self,
        query: str,
        actual_output: str,
        metric_score: float,
        metric_name: str,
        success: bool,
        datapoint_id: str,
        implementation_id: str,
        metrics_metadata: Any,
    ):
        """send test case data to the prod-data endpoint"""
        if not metrics_metadata:
            metrics_metadata = {}
        return self.post_request(
            endpoint="/v1/prod-data",
            body={
                "query": query,
                "actualOutput": actual_output,
                "defaultMetricScore": metric_score,
                "defaultMetricName": metric_name,
                "implementationId": implementation_id,
                "success": success,
                "metricsMetadata": metrics_metadata,
                "goldenId": datapoint_id,
            },
        )

    def add_golden(self, query: str, expected_output: str, is_synthetic: bool = False):
        return self.post_request(
            endpoint="/v1/golden",
            body={
                "query": query,
                "expectedOutput": expected_output,
                "isSynthetic": is_synthetic,
            },
        )

    def list_prod_data(self, implementation_id: str):
        return self.get_request(
            endpoint="/v1/prod-data?implementationId=" + implementation_id
        )

    def create_implementation(self, name: str, description: Optional[str] = None):

        body = {"name": name}
        if description:
            body["description"] = description
        return self.post_request(endpoint="/v1/implementation", body=body)

    def list_implementations(self):
        """
        Returns a list of implementations
        """
        return self.get_request(endpoint="/v1/implementation")



API_KEY_ENV: str = "CONFIDENT_AI_API_KEY"
LOG_TO_SERVER_ENV: str = "DO_NOT_SEND_TO_CONFIDENT_AI"
IMPLEMENTATION_ID_ENV: str = "CONFIDENT_AI_IMP_ID"
IMPLEMENTATION_ID_NAME: str = "CONFIDENT_AI_IMP_NAME"



def test_evaluation_dataset():
    from deepeval.dataset import EvaluationDataset

    csv_filename = "sample.csv"

    csv_file = """id,query,expected_output
    1,"Hello, world!","This is a greeting."
    2,"OpenAI GPT-3","A powerful language model."
    3,"CSV Example","Working with CSV data."
    4,"Python Programming","Coding in Python."
    5,"Data Science","Analyzing data."
    """

    with open(csv_filename, "w") as file:
        file.write(csv_file)

    dataset = EvaluationDataset.from_csv(
        csv_filename,
        query_column="query",
        expected_output_column="expected_output",
        id_column="id",
    )



"""Metric for bias classifier - using the same min score subtraction methodology as the toxic classifier 
Rationale for bias classifier is described here https://arxiv.org/pdf/2208.05777.pdf
1 - Not Biased
0 - Bias
"""

import warnings
import asyncio
from typing import Optional
from .metric import Metric
from ..singleton import Singleton


class UnBiasedMetric(Metric, metaclass=Singleton):
    def __init__(
        self, model_name: str = "original", minimum_score: float = 0.5
    ):  # see paper for rationale https://arxiv.org/pdf/2208.05777.pdf
        self.model_name = model_name
        self.minimum_score = minimum_score

    def __call__(self, output, expected_output, query: Optional[str] = "-"):
        score = self.measure(output, expected_output)
        success = score >= self.minimum_score
        asyncio.create_task(
            self._send_to_server(
                metric_score=score,
                metric_name=self.__name__,
                query=query,
                output=output,
                expected_output=expected_output,
                success=success,
            )
        )
        return score

    def measure(self, text: str):
        from Dbias.bias_classification import classifier

        warnings.warn(
            "Run `pip install deepeval[bias]`. If you have, please ignore this warning."
        )

        results = classifier(text)
        # sample output
        # [{'label': 'Biased', 'score': 0.9938021898269653}]
        if results[0]["label"] == "Biased":
            v = 0.5 - (results[0]["score"] / 2)
        else:
            # if it's unbiased - use normal score
            v = 0.5 + (results[0]["score"] / 2)

        self.success = False
        if v > self.minimum_score:
            self.success = True

        return v

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Bias Score"


def assert_unbiased(
    text: str, minimum_score: float = 0.5
):  # see paper for rationale https://arxiv.org/pdf/2208.05777.pdf
    metric = UnBiasedMetric(minimum_score=minimum_score)
    score = metric.measure(text)
    assert metric.is_successful(), f"Text is biased - got {score}"



import random
from .metric import Metric


class RandomMetric(Metric):
    def __init__(self, minimum_score: float = 0.5):
        self.minimum_score = minimum_score

    def measure(self, a, b):
        score = random.random()
        self.success = score >= self.minimum_score
        return score

    def is_successful(self):
        return self.success



class Singleton(type):
    """
    Singleton class for having a single instance of a class.
    This ensures that instances aren't created more than once.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]

    def __setattr__(cls, name, value):
        super().__setattr__(name, value)



# from deepeval.bulk_runner import BulkTestRunner, TestCase


# def generate_llm_output(input: str):
#     expected_output = "Our customer success phone line is 1200-231-231."
#     return expected_output


# def test_bulk_runner():
#     class BulkTester(BulkTestRunner):
#         @property
#         def bulk_test_cases(self):
#             return [
#                 TestCase(
#                     input="What is the customer success number",
#                     expected_output="1800-213-123",
#                 )
#             ]

#     tester = BulkTester()
#     tester.run(completion_fn=generate_llm_output)



import numpy as np
import os


def softmax(x):
    e_x = np.exp(x - np.max(x, axis=1, keepdims=True))
    return e_x / e_x.sum(axis=1, keepdims=True)


def cosine_similarity(vector_a, vector_b):
    dot_product = np.dot(vector_a, vector_b)
    norm_a = np.linalg.norm(vector_a)
    norm_b = np.linalg.norm(vector_b)
    similarity = dot_product / (norm_a * norm_b)
    return similarity


def chunk_text(text, chunk_size=20):
    words = text.split()
    chunks = [words[i : i + chunk_size] for i in range(0, len(words), chunk_size)]
    return chunks
